/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 539 $ $LastChangedDate: 2002-08-25 13:25:13 +0200 (Sun, 25 Aug 2002) $
 *
 * Bind - Name binder. 
 * 
 * In general it changes all occurrences of SimpleName[_] and 
 * QualifiedName[_] to Symbol[_], so next phases of compilation don't need 
 * to care about symbol tables, scopes and so forth.
 *
 * In addition we do some closure marking stuff here. Variables are
 * marked to be put into closure, closurized variables lists are created
 * for each function (including lambda functions), each reference to
 * symbol annotated with depth of walk in closure stack needed to get
 * to it.
 */

/* someday-TODO: check if it's still true ;)
 * Level stuff:
 *   fnc_level is level of function nesting for symbol.
 *     -1 for symbols from other modules
 *	0 for global symbols (in current module)
 *	1 for variables and parameters to 1st level functions
 *	2 ditto for 2nd level...
 *   scope_level is level of scope nesting of symbol, levels -2 up to
 *   0 are the same as for fnc_level, next define level of block nesting.
 *
 * Example	      // (fnc_level, scope_level)
 * int foo;	      // (0, 0)
 * int bar(	      // (0, 0)
 *     int x) {	      // (1, 1)
 *     int baz;	      // (1, 1)
 *     int f() {      // (1, 1)
 *	   int k;     // (2, 2)
 *     }
 *     { int z; }     // (1, 2)
 *  }
 */


// speed-TODO: use hashtables instead of lists for local symbols. 
// Hashtbl::copy() can take same time though, we might be better off
// doing some kind of copy_on_write(), or better yet cache few entries
// in list before doing a real copy (if it's needed, therefore there would
// be no Hashtbl::copy() for `{ int i; i = 5; }' or `{ x = 5; }'
// Note however, that Hashtbl::copy() has linear time, while now *searches*
// are linear... Ok, it's to be investigated.

open Ast;
open Util;
open Bind;

struct bind_ctx {
    int current_fnc_level;
    int current_scope_level;
    int current_loop_level;
    
    <var_symbol>List::t local_vars;
    <li_symbol>List::t loop_ids;

    <var_symbol>Hashtbl::t global_vars;
    <ty_symbol>Hashtbl::t tydefs;
    <sf_symbol>Hashtbl::t struct_fields;

    // array refers to function levels
    <<var_symbol>List::t>Array::t closures;

    <var_symbol>List::t formal_parms;

    <id>List::t open_modules;
  
    <int>Hashtbl::t referenced_modules;
    <int>Hashtbl::t used_modules;
    <int>Hashtbl::t used_exns;
    
    <id>List::t bound_tyvars;

    bool needs_closure;

    // variables bound by current pattern
    <var_symbol>List::t pattern_vars;

    bind_ctx parent;
}

bool done_init;
bind_ctx global_ctx;
id current_module;
bool doing_iface;

// reference_module() is used with Open and when types from module are mentioned
void reference_module(id m)
{
    if (!Hashtbl::mem(global_ctx.referenced_modules, m))
	Hashtbl::add(global_ctx.referenced_modules, m, 1);
}

// to use_module() means to need variable or function from it
void use_module(id m)
{
    if (!Hashtbl::mem(global_ctx.used_modules, m)) {
	Hashtbl::add(global_ctx.used_modules, m, 1);
	// used => referenced
	reference_module(m);
    }
}

*(bool (<'a>symbol)) sym_named(id x)
{
    return fun bool (<'a>symbol s) (s.name == x);
}

void init_binder(id mod_name, bool iface)
{
    global_ctx = ({
	current_fnc_level = 0,
	current_scope_level = 0,
	current_loop_level = 0,
    
	local_vars = List::Nil,
	loop_ids = List::Nil,
	open_modules = List::Nil,

	bound_tyvars = List::Nil,

	formal_parms = List::Nil,

	global_vars = Hashtbl::create(13),
	tydefs = Hashtbl::create(13),
	struct_fields = Hashtbl::create(13),

	closures = Array::make(10, List::Nil),

	used_exns = Hashtbl::create(13),
	used_modules = Hashtbl::create(13),
	referenced_modules = Hashtbl::create(13),

	needs_closure = true,

	pattern_vars = List::Nil,
	
	// dummy
	parent = global_ctx
    });

    // really we set it here
    global_ctx.parent = global_ctx;

    current_module = mod_name;
    doing_iface = iface;

    if (mod_name != "Std") {
	global_ctx.open_modules = List::cons1("Std");
	Module::open_module("Std");
 	reference_module("Std");
    }

    done_init = true;
}

void check_init()
{
    if (!done_init)
	ice("Bind::check_init(): didn't Bind::init_binder()!");
}

bind_ctx copy_context(bind_ctx parent)
{
    return ({
	current_scope_level = parent.current_scope_level + 1,
	parent = parent,
    
	// rest is simply copied:
	current_fnc_level = parent.current_fnc_level,
	current_loop_level = parent.current_loop_level,
	local_vars = parent.local_vars,
	global_vars = parent.global_vars,
	tydefs = parent.tydefs,
	loop_ids = parent.loop_ids,
	closures = parent.closures,
	struct_fields = parent.struct_fields,
	bound_tyvars = parent.bound_tyvars,
	formal_parms = parent.formal_parms,
	used_exns = parent.used_exns,
	used_modules = parent.used_modules,
	referenced_modules = parent.referenced_modules,
	open_modules = parent.open_modules,
	pattern_vars = parent.pattern_vars,
	needs_closure = false
    });
}

void check_if_in_current_module(bind_ctx c, <'a>symbol s)
{
    if (s.module != "" && s.module != current_module) {
	error("cannot define symbols from module `" + s.module +
	      "' in `" + current_module + "'");
	s.module = current_module;
    }
}

*[string, string] fetch_module_name(bind_ctx c, <'a>bindable src)
{
    switch src {
    case SimpleName[n]:
	// if in global scoping, must prepend current module
	// otherwise it is not necessary
	if (c.current_scope_level == 0) 
	    return [current_module, n];
	else
	    return ["", n];
	
    case QualifiedName[m, n]:
	return [m, n];
	
    case Symbol[s]:
	return ice("fetch_module_name(): already made");
    }
}

// We need to remove typevariables from upper-level function here. 
// I.e. in void f('a) { void g('a, 'b) {}}  for g it ft.parms should 
// be [b], not [a,b] as suggested by parser.
<id>List::t fixup_tyvars(bind_ctx c, <id>List::t tyvars)
{
    bool isnt_global_tyvar(id n)
    {
	return !List::exists(fun bool (id x) (x == n), c.bound_tyvars);
    }
    
    return List::filter(isnt_global_tyvar, tyvars);
    
}

// someday-TODO: make warning if new structure extends or narrows
// old one (i.e. there is no way one may make instance of it with
// ({ }))
void make_struct_field(bind_ctx c, ty_symbol tydef, <sf_symbol>field fld)
{
    push_location(fld.fld_loc);
    
    try {
	sf_symbol s;
	
	let [module, name] = fetch_module_name(c, fld.fld_symbol) in
	s = ({
	    module = module, 
	    name = name, 
	    stype = fld.fld_type,
	    sloc = top_location(),
	    a = ({sf_tydef = tydef})
	});
    
	check_if_in_current_module(c, s);

	Hashtbl::add(c.struct_fields, s.name, s);

	fld.fld_symbol = Symbol[s];
	
    } finally {
	pop_location();
    }
}

var_symbol make_var(bind_ctx c, ty t, <var_symbol>bindable src)
{
    string module, name, xc_name;

    [module, name] = fetch_module_name(c, src);

    if (module == "")
	xc_name = name;
    else
	xc_name = module + "$$" + name;

    var_symbol s = ({module = module, 
		     name = name, 
		     stype = t,
		     sloc = top_location(),
		     a = ({fnc_level = c.current_fnc_level,
			   scope_level = c.current_scope_level,
			   properties = List::Nil,
			   extern_name = xc_name
			  })
		    });
    
    check_if_in_current_module(c, s);

    List::iter((fun void (var_symbol x) {
	if (x.name == s.name) {
	    if (x.a.scope_level == s.a.scope_level) {
		error("redefinition of symbol `" + name + "'");
	    } else if (flags.wshadow) {
		warn("symbol `" + name + "' shadows other symbol");
		warnl(x.sloc, "this is the location of previous definition");
	    }
	}}), List::concat(Hashtbl::find_all(c.global_vars, s.name), 
			  c.local_vars));
    
    if (c.current_scope_level == 0)
	Hashtbl::add(c.global_vars, s.name, s);
    else
	c.local_vars = List::Cons[s, c.local_vars];

    return s;
}

void bind_tydef(bind_ctx c, ty_symbol tydef, ty t)
{
    void add_struct_field(<sf_symbol>field fld)
    {
	make_struct_field(c, tydef, fld);
    }
    
    void add_union_field(<var_symbol>field fld)
    {
	let s = make_var(c, fld.fld_type, fld.fld_symbol) in {
	    set_property(s, IsUnionField);
	    set_property(s, ParentUnion[tydef]);
	    fld.fld_symbol = Symbol[s];
	}
    }

    void f(ty tt)
    {
	switch tt {
	case Struct[flds]:
	    List::iter(add_struct_field, flds);
	    
	case OptStruct[flds]:
	    List::iter(add_struct_field, flds);

	case Union[flds]:
	    List::iter(add_union_field, flds);

	case _:
	    skip;
	}
    }

    type_iter(f, t);
}

void push_ty_symbol(bind_ctx c, ty_symbol tysym)
{
    if (tysym.module == "")
	tysym.module = current_module;

    check_if_in_current_module(c, tysym);

    void maybe_redef_error(ty_symbol x)
    {
	switch x.stype {
	case AbstractType:
	    if (List::length(x.a.tysym_tyvars) !=
		List::length(tysym.a.tysym_tyvars)) {
		error("different number of type arguments " +
		      "then in forward declaration");
		errorl(x.sloc, "here");
	    }

	case _:
	    error("redefinition of type `" + tysym.name + "'");
	    errorl(x.sloc, "first defined here");
	}
    }

    bool check_redef(ty_symbol x) 
    {
	maybe_redef_error(x);
	
	x.a.tysym_tyvars = tysym.a.tysym_tyvars;
	x.stype = tysym.stype;
	tysym = x;

	return true;
    }


    if (!doing_iface)
	try {
	    maybe_redef_error(Module::lookup_tysym(tysym.module,
				   tysym.name));
	} with { case Not_found: skip; }
    
    if (!List::exists(check_redef, Hashtbl::find_all(c.tydefs, tysym.name)))
	Hashtbl::add(c.tydefs, tysym.name, tysym);

    bind_tydef(c, tysym, tysym.stype);
}

<'a>symbol lookup_symbol(bind_ctx c, <<'a>symbol>bindable bound_name, 
			 *(<'a>symbol (id)) local_lookup,
			 *(<'a>symbol (id, id)) global_lookup)
{

    switch bound_name {
    case SimpleName[name]:
	try {
	    return local_lookup(name);
	} with {
	case Not_found:
	    <<'a>symbol>option ret = None;
	    string name_copy = name;
	    
	    bool lookup(id mod)
	    {
		try {
		    ret = Some[global_lookup(mod, name_copy)];
		    return true;
		} with {
		case Not_found:
		    return false;
		}
	    }
	    
	    if (List::exists(lookup, c.open_modules)) {
		switch ret {
		case Some[s]:
		    return s;
		case None:
		    return ice("Bind::lookup_symbol(): wtf?");
		}
	    } else {
		// sorry Winetou ;)
		raise Not_found;
	    }
	}
	
    case QualifiedName[m, n]:
	return global_lookup(m, n);
    
    case Symbol[s]:
	return s;
    }
}

ty_symbol lookup_tydef(bind_ctx c, <ty_symbol>bindable name)
{
    ty_symbol r;

    r = lookup_symbol(c, name, partial(Hashtbl::find, c.tydefs),
		      Module::lookup_tysym);
		      
    reference_module(r.module);
    
    return r;
}

var_symbol lookup_var(bind_ctx c, <var_symbol>bindable bname)
{
    var_symbol lookup(id name)
    {
	var_symbol lookup_in_lst(<var_symbol>List::t lst)
	{
	    switch lst {
	    case List::Cons[hd, tl]:
		if (hd.name == name)
		    return hd;
		else
		    return lookup_in_lst(tl);

	    case List::Nil:
		return Hashtbl::find(c.global_vars, name);
	    }
	}

	return lookup_in_lst(c.local_vars);
    }
    
    var_symbol s = lookup_symbol(c, bname, lookup, Module::lookup_var);

    if (has_property(s, IsUnionField)) {
	reference_module(s.module);
	if (has_property(s, IsException)) {
	    if (!Hashtbl::mem(global_ctx.used_exns, s.a.extern_name))
		Hashtbl::add(global_ctx.used_exns, s.a.extern_name, 1);
	}
    } else
	use_module(s.module);

    return s;
}

void bind_type(bind_ctx c, ty t)
{
    <ty>List::t extend_or_cut(int n, <ty>List::t lst)
    {
	if (n == 0)
	    return List::Nil;
	    
	switch lst {
	case List::Nil:
	    return List::Cons[AbstractType, extend_or_cut(n - 1, lst)];
	    
	case List::Cons[hd, tl]:
	    return List::Cons[hd, extend_or_cut(n - 1, tl)];
	}
    }
    
    void check_parms(nt_record nt)
    {
	let ts = get_binding(nt.tydef) in
	let act_len = List::length(nt.parms) in
	let frm_len = List::length(ts.a.tysym_tyvars) in
	if (act_len != frm_len) {
	    error("number of type arguments of `" + 
		  symbol_name(ts) + "' doesn't match definition");
	    errorl(ts.sloc, "(definition was here)");
	    nt.parms = extend_or_cut(frm_len, nt.parms);
	}
    }
    
    void f(ty tt)
    {
	switch tt {
	case NamedType[nt]:
	    try {
		let tydef = lookup_tydef(c, nt.tydef) in {
		    nt.tydef = Symbol[tydef];
		    reference_module(tydef.module);
		}
		check_parms(nt);
	    } with {
	    case Not_found:
		error("unbound type `" + qualified_name(nt.tydef) + "'");
		let [m, n] = fetch_module_name(c, nt.tydef) in
		let artificial_tydef = ({name = n, module = m,
					 stype = AbstractType,
					 sloc = top_location(),
					 a = ({tysym_tyvars = List::Nil})
					}) in
		nt.tydef = Symbol[artificial_tydef];
	    }

	case Func[ft]:
	    ft.tyvars = fixup_tyvars(c, ft.tyvars);

	case _:
	    skip;
	}
    }

    type_iter(f, t);
}

var_symbol enter_vardef(bind_ctx c, vardef v)
{
    let s = make_var(c, v.var_type, v.var_name) in {
	v.var_name = Symbol[s];
	return s;
    }
}

var_symbol bind_vardef(bind_ctx c, vardef v)
{
    bind_type(c, v.var_type);
    return enter_vardef(c, v);
}

bind_ctx make_fn_ctx(bind_ctx c, ty fn_type)
{
    bind_ctx fn_ctx = copy_context(c);
    fn_ctx.formal_parms = List::Nil;
    fn_ctx.current_fnc_level++;
    Array::set(fn_ctx.closures, fn_ctx.current_fnc_level, List::Nil);

    void enter_parm(*[ty, <id>option] p)
    {
	switch p {
	case [t, Some[n]]:
	    let v = ({var_name = SimpleName[n], var_type = t}) in
	    let s = bind_vardef(fn_ctx, v) in {
		set_property(s, IsLvalue);
		set_property(s, IsInitialized);
		fn_ctx.formal_parms = List::Cons[s, fn_ctx.formal_parms];
	    }
	case [_, None]:
	    skip;
	}
    }

    let ft = get_ft_record(fn_type) in {
	bind_type(fn_ctx, fn_type);
	fn_ctx.bound_tyvars = List::concat(ft.tyvars, fn_ctx.bound_tyvars);
	List::iter(enter_parm, ft.parms);
    }

    fn_ctx.formal_parms = List::rev(fn_ctx.formal_parms);

    return fn_ctx;
}

bind_ctx make_ty_ctx(bind_ctx c, ty_symbol ts)
{
    bind_ctx ty_ctx = copy_context(c);

    ty_ctx.current_scope_level = c.current_scope_level;
    ty_ctx.bound_tyvars = List::concat(ts.a.tysym_tyvars, c.bound_tyvars);

    return ty_ctx;
}

void bind_lambda(bind_ctx c, lambda_record l)
{
    bind_ctx fn_ctx = make_fn_ctx(c, l.lambda_type);
    
    l.lambda_parms = fn_ctx.formal_parms;

    if (fn_ctx.current_fnc_level + 2 >= Array::length(c.closures))
	Array::resize(c.closures, fn_ctx.current_fnc_level * 3, List::Nil);
	
    bind_stmt(fn_ctx, l.lambda_body);
	
    l.lambda_closure_vars = 
	    Array::get(fn_ctx.closures, fn_ctx.current_fnc_level);
    l.lambda_needs_closure = fn_ctx.needs_closure;

    // sanity checks
    switch [l.lambda_needs_closure, l.lambda_closure_vars] {
    case [false, List::Cons[_, _]]:
	ice("Bind::bind_lambda(): needs no closure, but has closure vars");

    case [true, List::Nil]:
	if (fn_ctx.current_fnc_level == 1)
	    ice("Bind::bind_lambda(): needs no closure, but need one");

    case _:
	skip;
    }
}

<sf_symbol>List::t lookup_struct_field(bind_ctx c, <sf_symbol>bindable name)
{
    <sf_symbol>List::t find_all(<sf_symbol>List::t acc, 
			        <id>List::t modules, id n)
    {
	switch modules {
	case List::Cons[m, tl]:
	    let th = Module::lookup_struct_field(m, n) in
	    return find_all(List::concat(th, acc), tl, n);
	case List::Nil:
	    return acc;
	}
    }
    
    switch name {
    case Symbol[s]:
	return List::cons1(s);
	
    case QualifiedName[m, n]:
	reference_module(m);
	return Module::lookup_struct_field(m, n);
    
    case SimpleName[n]:
	let here = Hashtbl::find_all(c.struct_fields, n) in
	let there = find_all(List::Nil, c.open_modules, n) in
	return List::concat(here, there);
    }
}

void bind_struct_expr(bind_ctx c, <sini_expr>List::t ses)
{
    <ty_symbol>List::t possible_tydefs = List::Nil;

    <ty_symbol>List::t lookup_tydefs(<sf_symbol>bindable name)
    {
	return List::map(fun ty_symbol (sf_symbol s) (s.a.sf_tydef), 
			 lookup_struct_field(c, name));
    }
    
    bool check_if_possible(ty_symbol t)
    {
	return List::exists(fun bool (ty_symbol p) 
				(p.name == t.name && p.module == t.module),
			    possible_tydefs);
    }
    
    void bind_one(sini_expr se)
    {
	push_location(se.sx_expr.loc);
	try {
	    bind_expr(c, se.sx_expr);
	    
	    let new = List::filter(check_if_possible, 
				   lookup_tydefs(se.sx_symbol)) in
	    if (List::length(new) == 0) {
		let names = 
		  String::concat(" nor ", 
				 List::map(symbol_name, possible_tydefs)) in
		error("field `" + qualified_name(se.sx_symbol) + 
		      "' is not part of " + names);
	    } else {
		possible_tydefs = new;
	    }
	} finally {
	    pop_location();
	}
    }

    switch ses {
    case List::Cons[hd, _]:
	possible_tydefs = lookup_tydefs(hd.sx_symbol);
	
    case List::Nil:
	ice("Bind::bind_struct_expr(): empty");
    }
    
    List::iter(bind_one, ses);

    if (List::length(possible_tydefs) > 1) {
	let names = String::concat(", ", List::map(symbol_name,
				   possible_tydefs)) in
	error("more then one type possible for expression: " + names);
    }

    switch possible_tydefs {
    case List::Cons[ts, _]:
	<<sf_symbol>field>List::t content;
	
	switch ts.stype {
	case Struct[x]:
	    content = x;
	    
	case OptStruct[x]:
	    content = x;

	case _:
	    ice("Bind::bind_struct_expr(): non struct");
	    return;
	}
	
	void set_sym(sini_expr se)
	{
	    bool check(<sf_symbol>field f)
	    {
		return simple_name(f.fld_symbol) == simple_name(se.sx_symbol);
	    }
	    
	    let f = List::find(check, content) in
	    se.sx_symbol = Symbol[get_binding(f.fld_symbol)];
	}
	
	try {
	    List::iter(set_sym, ses);
	} with {
	case Not_found:
	    ice("Bind::bind_struct_expr(): not found");
	}


    case List::Nil:
	raise Choke;	// sorry, cannot bind anything
    }
}

void bind_ref(bind_ctx c, nl_expr ex, ref_record rr)
{
    try {
	var_symbol s = lookup_var(c, rr.sym);

	if (has_property(s, IsUnionField)) {
	    // this is sth like List::Nil, we need to change
	    // this expression node to reflect this fact.
	    let void_const = ({loc = top_location(),
			       a = ({raw = VoidConst,
				     expr_type = None})}) in
	    ex.raw = UnionExpr[({nx_symbol = Symbol[s], nx_expr = void_const})];
	    return;
	}

	rr.sym = Symbol[s];
	
	if (s.a.fnc_level > 0)
	    rr.level = c.current_fnc_level - s.a.fnc_level;
	else
	    rr.level = 0;
	    
	if (rr.level > 0 && !has_property(s, IsInClosure)) {
	    set_property(s, IsInClosure);
	    let lst = Array::get(c.closures, s.a.fnc_level) in
	    Array::set(c.closures, s.a.fnc_level, List::Cons[s, lst]);

	    for (bind_ctx p = c; p.current_fnc_level >= s.a.fnc_level; p = p.parent)
		if (p.current_fnc_level < c.current_fnc_level)
		    p.needs_closure = true;
	}
    } with {
    case Not_found:
	error("unbound variable `" + qualified_name(rr.sym) + "'");
	raise Choke;
    }
}

var_symbol bind_union_tag(bind_ctx c, <var_symbol>bindable sym)
{
    try {
	let s = lookup_var(c, sym) in 
	if (!has_property(s, IsUnionField)) {
	    error("`" + qualified_name(sym) +
		  "' is not part of any union");
	    raise Choke;
	} else {
	    return s;
	}
    } with {
    case Not_found:
	error("unbound union field `" + qualified_name(sym) + "'");
	raise Choke;
    }
}

void bind_raw_expr(bind_ctx c, nl_expr ex)
{
    switch ex.raw {
    case SimpleBinOp[_, e1, e2]:
	bind_expr(c, e1);
	bind_expr(c, e2);
	
    case SimpleUnOp[_, e]:
	bind_expr(c, e);
    
    case Ref[rr]:
	bind_ref(c, ex, rr);

    case Call[e, el]:
	bind_expr(c, e);
	List::iter(partial(bind_expr, c), el);

    case TupleExpr[el]:
	List::iter(partial(bind_expr, c), el);

    case UnionExpr[u]:
	bind_expr(c, u.nx_expr);
	u.nx_symbol = Symbol[bind_union_tag(c, u.nx_symbol)];
    
    case Lambda[l]:
	bind_lambda(c, l);

    case StructExpr[s]:
	bind_struct_expr(c, s);

    case DotExpr[de]:
	// we cannot set nx_symbol here, it has to wait for typechecker
	bind_expr(c, de.sx_expr);
    
    // contants arn't really interesting in this context ;)
    case BoolConst[_]:	    skip;
    case IntConst[_]:	    skip;
    case FloatConst[_]:	    skip;
    case StringConst[_]:    skip;
    case VoidConst:	    skip;
    case NullConst:	    skip;
    }

}

void bind_expr(bind_ctx c, expr e)
{
    locate(partial(bind_raw_expr, c), e);
}

bool is_capitalized(string s)
{
    return int_of_char(String::uppercase_char(String::get(s, 0))) == 
	   int_of_char(String::get(s, 0));
}

void bind_pattern_int(bind_ctx c, pattern p)
{
    switch p.a {
    case VoidPattern:
	skip;
	
    case VarPattern[v]:
	let s = bind_vardef(c, v) in {
	    if (flags.wcapitalization && 
		is_capitalized(simple_name(v.var_name)))
		warn("capitalized pattern variable `" +
		     simple_name(v.var_name) + "'");
	    set_property(s, IsInPattern);
	    c.pattern_vars = List::Cons[s, c.pattern_vars];
	}
    
    case ExprPattern[e]:
	/* 
	 * ExprPattern[...] is generated for `Nil', `List::Nil', 
	 * `x' and `y' where Nil and List::Nil are union members, x 
	 * is some bound variable and y is free variable. We fixup 
	 * it here. Note that type of VarPattern[..] vardef is set
	 * to AbstractType. This is because we cannot tell the
	 * actual type here (at least not in all cases).
	 */
	switch e.a.raw {
	case Ref[rr]:
	    try {
		let s = lookup_var(c, rr.sym) in
		if (has_property(s, IsUnionField)) {
		    let np = ({np_symbol = Symbol[s],
			       np_pattern = ({loc = top_location(), 
					      a = VoidPattern[]})
			      }) in
		    p.a = UnionPattern[np];
		    return;
		} else {
		    error("variable " + simple_name(rr.sym) + 
			  " is bound but isn't union field");
		}
	    } with {
	    case Not_found:
		let xt = ({t = None, name = genname("_patvar")}) in
		p.a = VarPattern[({var_name = rr.sym, 
				   var_type = XTyVar[xt]})];
		bind_pattern_int(c, p);
		return;
	    }
	case _:
	    skip;
	}
	bind_expr(c, e);

    // Fixup 0 and 1-long tuple patterns
    case TuplePattern[List::Nil]:
	p.a = VoidPattern;

    case TuplePattern[List::Cons[x, List::Nil]]:
	p.loc = x.loc;
	p.a = x.a;
	bind_pattern_int(c, p);
	
    case TuplePattern[lst]:
	List::iter(partial(bind_pattern, c), lst);
    
    case UnionPattern[np]:
	np.np_symbol = Symbol[bind_union_tag(c, np.np_symbol)];
	bind_pattern_int(c, np.np_pattern);
    }
}

void bind_pattern(bind_ctx c, pattern p)
{
    push_location(p.loc);
    try {
	bind_pattern_int(c, p);
    } finally {
	pop_location();
    }
}

struct pattern_var_info {
    var_symbol primary_symbol;
    int occurred_cnt;
}

void bind_case_clause(bind_ctx c, case_clause cc)
{
    <pattern_var_info>Hashtbl::t vars = Hashtbl::create(13);
    bool first = true;
    int ncases = List::length(cc.cases);

    <var_symbol>List::t bind_one_case_1(*[pattern, expr] oc)
    {
	void add_to_hash(var_symbol s)
	{
	    if (first) {
		Hashtbl::add(vars, s.name,
			     ({primary_symbol = s, occurred_cnt = 1}));
	    } else {
		try {
		    let i = Hashtbl::find(vars, s.name) in
		    i.occurred_cnt++;
		} with { case Not_found: skip; }
	    }
	}
	
	let [pat, _] = oc in
	let ctx = copy_context(c) in {
	    if (List::length(ctx.pattern_vars) != 0)
		ice("Bind::bind_case_clause(): pattern_vars not clear");
	    bind_pattern(ctx, pat);
	    List::iter(add_to_hash, ctx.pattern_vars);
	    first = false;
	    return ctx.pattern_vars;
	}
    }

    void bind_one_case_2(*[pattern, expr] oc, <var_symbol>List::t patvars)
    {
	var_symbol maybe_change(var_symbol s)
	{
	    try {
		let i = Hashtbl::find(vars, s.name) in
		if (i.occurred_cnt == ncases)
		    return i.primary_symbol;
	    } with { case Not_found: skip; }

	    if (!has_property(s, IsGuardLocal))
		set_property(s, IsGuardLocal);
	
	    return s;
	}

	void change_pattern(pattern p)
	{
	    switch p.a {
		case VarPattern[v]:
		    let s = get_binding(v.var_name) in {
			v.var_name = Symbol[maybe_change(s)];
			v.var_type = s.stype;
		    }
		
		case VoidPattern:	skip;
		case ExprPattern[_]:    skip;
		
		case TuplePattern[lst]:
		    List::iter(change_pattern, lst);
		
		case UnionPattern[np]:
		    change_pattern(np.np_pattern);
	    }
	}

	let [pat, cond] = oc in
	let ctx = copy_context(c) in {
	    ctx.local_vars = List::concat(List::map(maybe_change, patvars),
					  ctx.local_vars);
	    bind_expr(ctx, cond);
	    change_pattern(pat);
	}
    }

    let patvars_lists = List::map(bind_one_case_1, cc.cases) in
    List::iter2(bind_one_case_2, cc.cases, patvars_lists);
    
    let ctx = copy_context(c) in {
	void maybe_push(string, pattern_var_info i)
	{
	    if (i.occurred_cnt == ncases) {
		ctx.local_vars = List::Cons[i.primary_symbol, ctx.local_vars];
		cc.patvars = List::Cons[i.primary_symbol, cc.patvars];
	    }
	}
	
	Hashtbl::iter(vars, maybe_push);

	switch cc.body.a {
	case Block[List::Nil]:
	    ice("Bind::bind_case_clause(): empty case body");
	case _:
	    bind_stmt(ctx, cc.body);
	}
    }
}

void open_module(bind_ctx c, id mod)
{
    if (!List::exists(fun bool (id x) (x == mod), c.open_modules))
	try {
	    // error-check
	    Module::open_module(mod);
	    c.open_modules = List::Cons[mod, c.open_modules];
	    reference_module(mod);
	} with {
	case Choke:
	    // just don't open it.
	    skip;
	}
}

void bind_loop(bind_ctx c, loop_record l)
{
    string name;

    switch l.ll_sym {
    case Symbol[_]:
	ice("Bind::bind_loop(): bound");
	return;
	
    case QualifiedName[_, _]:
	ice("Bind::bind_loop(): qn");
	return;
    
    case SimpleName[n]:
	name = n;
    }
    
    if (name != "" && flags.wshadow && 
	List::exists(sym_named(name), c.loop_ids))
	warn("loop label `" + name + "' shadows another loop label");
    
    let loop_ctx = copy_context(c) in {
	loop_ctx.current_loop_level = c.current_loop_level + 1;
	
	let sym = ({module = "", 
		    name = name, 
		    stype = AbstractType,
		    sloc = top_location(), 
		    a = ({loop_level = loop_ctx.current_loop_level})
		   }) in 
	{
	    loop_ctx.loop_ids = List::Cons[sym, loop_ctx.loop_ids];
	    l.ll_sym = Symbol[sym];
	}
    
	bind_stmt(copy_context(loop_ctx), l.ll_body);
	bind_stmt(copy_context(loop_ctx), l.ll_inc);
    }
}

void bind_break(bind_ctx c, break_record b)
{
    string name;

    switch b.sym {
    case Symbol[_]:
	ice("Bind::bind_break(): bound");
	return;
	
    case QualifiedName[_, _]:
	ice("Bind::bind_break(): qn");
	return;
    
    case SimpleName[n]:
	name = n;
    }
    
    switch c.loop_ids {
    case List::Cons[hd, _]:
	if (name == "") {
	    b.sym = Symbol[hd];
	    if (hd.name != "" && flags.wlabeled) {
		warn("break without a label in labeled (with `" + hd.name +
		     "') loop");
	    }
	} else
	    try {
		let s = List::find(sym_named(name), c.loop_ids) in
		b.sym = Symbol[s];
	    } with {
	    case Not_found:
		error("unbound loop label `" + name + "'");
		b.sym = Symbol[hd];
	    }
    
    case List::Nil:
	error("break outside loop");
    }
}

void bind_let_pattern(bind_ctx c, pattern pat)
{
    if (List::length(c.pattern_vars) != 0)
	ice("Bind::bind_let_pattern(): pattern_vars not clear");
	
    bind_pattern(c, pat);
    
    List::iter(fun void (var_symbol s) {set_property(s, IsLetVariable);},
	       c.pattern_vars);

    c.pattern_vars = List::Nil;
}

void bind_nl_stmt(bind_ctx c, nl_stmt r)
{
    switch r {
    case Skip:
	skip;
    
    case Expr[e]:
	bind_expr(c, e);
    
    case Return[e]:
	bind_expr(c, e);

    case VarDef[v]:
	let s = bind_vardef(c, v) in
	set_property(s, IsLvalue);
	
    case Block[lst]:
	let block_ctx = copy_context(c) in
	List::iter(partial(bind_stmt, block_ctx), lst);

    case Pair[s1, s2]:
	bind_stmt(c, s1);
	bind_stmt(c, s2);
    
    case Switch[e, clauses]:
	bind_expr(c, e);
	List::iter(partial(bind_case_clause, c), clauses);
	
    case Let[assigns, s]:
	let ctx = copy_context(c) in {
	    List::iter(partial(bind_expr, ctx), List::map(snd, assigns));
	    List::iter(partial(bind_let_pattern, ctx), List::map(fst, assigns));
	    bind_stmt(ctx, s);
	}
    
    case If[e, s1, s2]:
	bind_expr(c, e);
	bind_stmt(copy_context(c), s1);
	bind_stmt(copy_context(c), s2);
    
    case Loop[l]:
	bind_loop(c, l);
    
    case Break[b]:
	bind_break(c, b);
	
    case BreakIfNot[e]:
	bind_expr(c, e);
    
    case Open[m]:
	open_module(c, m);
    
    case Try[tr]:
	bind_stmt(copy_context(c), tr.try_body);
	List::iter(partial(bind_case_clause, c), tr.try_with);
	bind_stmt(copy_context(c), tr.try_finally);
    
    case Raise[e]:
	bind_expr(c, e);
    }
}

void bind_stmt(bind_ctx c, stmt s)
{
    locate(partial(bind_nl_stmt, c), s);
}

<id>List::t simple_fv(ty t)
{
    <int>Hashtbl::t h = Hashtbl::create(13);

    void f(ty tt)
    {
	switch tt {
	case TyVar[x]:
	    if (!Hashtbl::mem(h, x))
		Hashtbl::add(h, x, 1);
	case _:
	    skip;
	}
    }

    type_iter(f, t);

    <id>List::t ret = List::Nil;
    Hashtbl::iter(h, (fun void (string x, int dummy) 
			    {ret = List::Cons[x, ret];}));
    
    return ret;
}

void iface_check(string m)
{
    if (doing_iface)
	error(m + " not allowed in interface");
}

// enter top-level symbol into binding context
void enter_nl_decl(nl_decl d)
{
    check_init();
    switch d {
    case FuncDecl[fn]:
	iface_check("function definition");
	let s = make_var(global_ctx, fn.l.lambda_type, fn.fn_sym) in {
	    fn.fn_sym = Symbol[s];
	}
	
    case GlobalDecl[v]:
	let s = enter_vardef(global_ctx, v) in {
	    set_property(s, IsLvalue);
	    set_property(s, IsGlobalVar);
	}
    
    case ExternC[v]:
	let s = enter_vardef(global_ctx, v) in {
	    s.a.extern_name = s.name;
	    set_property(s, IsExternC);
	    set_property(s, IsLvalue);
	}
    
    case FuncProto[v]:
	enter_vardef(global_ctx, v);

    case XCFuncProto[xc_name, v]:
	let s = enter_vardef(global_ctx, v) in {
	    set_property(s, IsExternC);
	    s.a.extern_name = xc_name;
	}
	
    case TopLevelOpen[m]:
	skip;
    
    case TyDef[tysym]:
	push_ty_symbol(global_ctx, tysym);
    
    case ExceptionDecl[v]:
	void poly_check(ty t)
	{
	    switch t {
	    case TyVar[_]:
		error("exception types cannot be polymorphic");
	    case _:
		skip;
	    }
	}

	type_iter(poly_check, v.var_type);
	let s = enter_vardef(global_ctx, v) in {
	    set_property(s, IsException);
	    set_property(s, IsUnionField);
	}
    
    case DummyDecl:
	skip;
    
    case SectionDecl[i, _]:
	iface_check("section definition");
	switch i {
	case "init":	skip;
	case "fini":	skip;
	case _:
	    error("unknown section `" + i + "'");
	}
    }
}

void enter_decl(decl d)
{
    locate(enter_nl_decl, d);
}

void bind_sym_type(bind_ctx c, <<'a>symbol>bindable sym)
{
    switch sym {
    case Symbol[s]:
	bind_type(c, s.stype);
    
    case _:
	ice("bind_sym_type(): not bound");
    }
}

void bind_nl_decl(nl_decl d)
{
    check_init();
    switch d {
    case TopLevelOpen[m]:
	open_module(global_ctx, m);
    
    case FuncDecl[fn]:
	bind_lambda(global_ctx, fn.l);
    
    case SectionDecl[_, l]:
	bind_lambda(global_ctx, l);
    
    // here just need to bind type
    case GlobalDecl[v]:
	bind_sym_type(global_ctx, v.var_name);
    
    case ExternC[v]:
	bind_sym_type(global_ctx, v.var_name);
    
    case FuncProto[v]:
	bind_sym_type(global_ctx, v.var_name);

    case XCFuncProto[_, v]:
	bind_sym_type(global_ctx, v.var_name);
	
    case TyDef[tysym]:
	bind_type(make_ty_ctx(global_ctx, tysym), tysym.stype);
    
    case ExceptionDecl[v]:
	bind_sym_type(global_ctx, v.var_name);

    // ble..
    case DummyDecl:
	skip;
    }
}

void bind_decl(decl d)
{
    locate(bind_nl_decl, d);
}

<id>List::t get_modules(<int>Hashtbl::t h)
{
    <id>List::t ret = List::Nil;
    
    void add(string name, int dummy)
    {
	if (name != current_module && name != "")
	    ret = List::Cons[name, ret];
    }
    
    Hashtbl::iter(h, add);

    return ret;
}

<id>List::t used_modules()
{
    check_init();
    return get_modules(global_ctx.used_modules);
}

<id>List::t referenced_modules()
{
    check_init();
    return get_modules(global_ctx.referenced_modules);
}

<id>List::t used_exns()
{
    check_init();
    return get_modules(global_ctx.used_exns);
}

void save_module(string filename)
{
    check_init();

    Module::module m = 
	   ({Module::mod_name = current_module,
	     Module::mod_under_construction = false,
	     Module::mod_vars = global_ctx.global_vars,
	     Module::mod_struct_fields = global_ctx.struct_fields,
	     Module::mod_tydefs = global_ctx.tydefs,
	     Module::mod_depends_on = referenced_modules()
	    });
    Module::save_module(m, filename);
}

// vim:sts=4:sw=4
